############################################
# imports
############################################

import joblib
import pandas as pd

from fgclustering.forest_guided_clustering import *


############################################
# Tests
############################################

def test_forest_guided_clustering():

    data_breast_cancer = pd.read_csv('./data/data_breast_cancer.csv')
    rf = joblib.load(open('./data/random_forest_breat_cancer.joblib', 'rb'))
    
    expected_output = 2
    
    # initialize and run fgclustering object
    fgc = FgClustering(model=rf, data=data_breast_cancer, target_column='target')
    fgc.run(max_K = 4, bootstraps_JI = 30, max_iter_clustering = 100, discart_value_JI = 0.6)

    # obtain optimal number of clusters and vector that contains the cluster label of each data point
    result = fgc.k
    
    assert result == expected_output






############################################
# imports
############################################

import numpy as np
from tqdm import tqdm

from sklearn_extra.cluster import KMedoids

import fgclustering.statistics as statistics

############################################
# Optimize number of clusters k
############################################


def _compute_jaccard_matrix(clusters, indices_bootstrap_clusters, indices_original_clusters):
    '''Compute Jaccard Index between all possible cluster combinations of original vs bootstrapped clustering.

    :param clusters: Clustering labels.
    :type clusters: numpy.ndarray
    :param indices_bootstrap_clusters: Dictionary with cluster labels as keys and index of instances that 
        belong to the respective cluster as labels for boostrapped clustering.
    :type indices_bootstrap_clusters: dict
    :param indices_original_clusters: Dictionary with cluster labels as keys and index of instances that 
        belong to the respective cluster as labels for original clustering.
    :type indices_original_clusters: dict
    :return: Jaccard Index for all cluster combinations.
    :rtype: numpy.ndarray
    '''
    indices_bootstrap_all = np.unique([index for i, cluster_bootstrap in enumerate(clusters) for index in indices_bootstrap_clusters[cluster_bootstrap]])
    jaccard_matrix = np.zeros([len(clusters), len(clusters)])

    for i, cluster_original in enumerate(clusters):
        for j, cluster_bootstrap in enumerate(clusters):
            indices_bootstrap = indices_bootstrap_clusters[cluster_bootstrap]
            indices_original = indices_original_clusters[cluster_original]
            
            # only compute overlap for instances that were in the whole bootstrap sample
            indices_original = indices_original.intersection(indices_bootstrap_all)

            intersection = indices_original.intersection(indices_bootstrap)
            union = indices_original.union(indices_bootstrap)
            
            jaccard_matrix[i,j] = len(intersection)/len(union)
            
    return jaccard_matrix


def _bootstrap_matrix(M):
    '''Create a bootstrap from the original matrix.

    :param M: Original matrix.
    :type M: pandas.DataFrame
    :return: M_bootstrapped: ootstrapped matrix; 
        mapping_bootstrapped_indices_to_original_indices: mapping from bootstrapped to original indices.
    :rtype: pandas.DataFrame, dict
    '''    
    lm = len(M)
    bootstrapped_samples = np.random.choice(np.arange(lm), lm)
    M_bootstrapped = M[:,bootstrapped_samples][bootstrapped_samples,:]
    
    mapping_bootstrapped_indices_to_original_indices = {bootstrapped : original for bootstrapped, original in enumerate(bootstrapped_samples)}
    
    return M_bootstrapped, mapping_bootstrapped_indices_to_original_indices


def _translate_cluster_labels_to_dictionary_of_index_sets_per_cluster(labels, mapping = False):
    '''Create dictionary that maps indices to cluster labels. 

    :param labels: Clustering labels.
    :type labels: numpy.ndarray
    :param mapping: Mapping of bootstrapped to original indices, defaults to False
    :type mapping: bool, optional
    :return: Dictionary with cluster labels as keys and index of instances that belong to the respective cluster as labels.
    :rtype: dict
    '''
    clusters = np.unique(labels)
    number_datapoints = len(labels)
    index_vector = np.arange(number_datapoints)
    
    indices_clusters = {}
    for cluster in clusters:
        indices = set(index_vector[labels == cluster])
        if mapping is not False:
            #translate from the bootstrapped indices to the original naming of the indices
            indices = set([mapping[index] for index in indices])

        indices_clusters[cluster] = indices

    return indices_clusters


def _compute_stability_indices(distance_matrix, cluster_method, bootstraps, random_state):
    '''Compute stability of each cluster via Jaccard Index of bootstraped vs original clustering.

    :param distance_matrix: Proximity matrix of Random Forest model.
    :type distance_matrix: pandas.DataFrame
    :param cluster_method: Lambda function wrapping the k-mediods clustering function.
    :type cluster_method: object
    :param bootstraps: Number of bootstraps to compute the Jaccard Index, defaults to 300
    :type bootstraps: int
    :param random_state: Seed number for random state, defaults to 42
    :type random_state: int
    :return: Dictionary with cluster labels as keys and Jaccard Indices as values.
    :rtype: dict
    '''
    np.random.seed = random_state
    
    matrix_shape = distance_matrix.shape
    assert len(matrix_shape) == 2, "error distance_matrix is not a matrix"
    assert matrix_shape[0] == matrix_shape[1], "error distance matrix is not square"
    
    labels = cluster_method(distance_matrix)
    clusters = np.unique(labels)
    number_datapoints = len(labels)
    index_vector = np.arange(number_datapoints)
    
    indices_original_clusters = _translate_cluster_labels_to_dictionary_of_index_sets_per_cluster(labels)
    index_per_cluster = {cluster: 0 for cluster in clusters}
    
    for i in range(bootstraps):
        bootstrapped_distance_matrix, mapping_bootstrapped_indices_to_original_indices = _bootstrap_matrix(distance_matrix)
        bootstrapped_labels = cluster_method(bootstrapped_distance_matrix)
        
        # now compute the indices for the different clusters
        indices_bootstrap_clusters = _translate_cluster_labels_to_dictionary_of_index_sets_per_cluster(bootstrapped_labels, mapping = mapping_bootstrapped_indices_to_original_indices)
        jaccard_matrix = _compute_jaccard_matrix(clusters, indices_bootstrap_clusters, indices_original_clusters)
        
        # compute optimal jaccard index for each cluster -> choose maximum possible jaccard index first
        for cluster_round in range(len(jaccard_matrix)):
            best_index = jaccard_matrix.max(axis=1).max()       
            original_cluster_number = jaccard_matrix.max(axis=1).argmax()
            bootstrapped_cluster_number = jaccard_matrix[original_cluster_number].argmax()
            jaccard_matrix[original_cluster_number] = -np.inf
            jaccard_matrix[:,bootstrapped_cluster_number] = -np.inf

            original_cluster = clusters[original_cluster_number]
            index_per_cluster[original_cluster] += best_index
                                    
    # normalize
    index_per_cluster = {cluster: index_per_cluster[cluster]/bootstraps for cluster in clusters}
        
    return index_per_cluster
    
    
def optimizeK(distance_matrix, y, max_K, bootstraps, max_iter_clustering, discart_value, method, random_state):
    '''Compute the optimal number of clusters for k-medoids clustering (trade-off between cluster purity and cluster stability). 

    :param distance_matrix: Proximity matrix of Random Forest model.
    :type distance_matrix: pandas.DataFrame
    :param y: Target column.
    :type y: pandas.Series
    :param max_K: Maximum number of clusters for cluster score computation, defaults to 6
    :type max_K: int
    :param bootstraps: Number of bootstraps to compute the Jaccard Index, defaults to 300
    :type bootstraps: int
    :param max_iter_clustering: Number of iterations for k-medoids clustering, defaults to 500
    :type max_iter_clustering: int
    :param discart_value: Minimum Jaccard Index for cluster stability, defaults to 0.6
    :type discart_value: float
    :param method: Model type of Random Forest model: classifier or regression.
    :type method: str
    :param random_state: Seed number for random state, defaults to 42
    :type random_state: int
    :return: Optimal number of clusters.
    :rtype: int
    '''    
    score_min = np.inf
    optimal_k = 1
    
    for k in tqdm(range(2, max_K)):
        #compute clusters        
        cluster_method = lambda X: KMedoids(n_clusters=k, random_state=random_state, init = 'build', method = "pam", max_iter=max_iter_clustering).fit(X).labels_
        labels = cluster_method(distance_matrix)

        # compute jaccard indices
        index_per_cluster = _compute_stability_indices(distance_matrix, cluster_method, bootstraps, random_state)
        min_index = min([index_per_cluster[cluster] for cluster in index_per_cluster.keys()])
        
        # only continue if jaccard indices are all larger 0.6 (thus all clusters are stable)
        print('For number of cluster {} the Jaccard Index is {}'.format(k, min_index))
        if min_index > discart_value:
            if method == "classifier":
                # compute balanced purities
                score = statistics.compute_balanced_average_impurity(y, labels)
            elif method == "regression":
                # compute the total within cluster variation
                score = statistics.compute_total_within_cluster_variation(y, labels)
            if score<score_min:
                optimal_k = k
                score_min = score
            print('For number of cluster {} the score is {}'.format(k,score))
        else:
            print('Clustering is instable, no score computed!')
            
    print(f"The optimal number of clusters is {optimal_k}")

    return optimal_k




############################################
# imports
############################################

import joblib
import pandas as pd


from fgclustering.utils import *


############################################
# Tests
############################################


def test_proximityMatrix():  
    
    data_breast_cancer = pd.read_csv('./data/data_breast_cancer.csv')
    model = joblib.load(open('./data/random_forest_breat_cancer.joblib', 'rb'))
    X = data_breast_cancer.drop(columns=['target']).to_numpy()
    
    result = proximityMatrix(model, X)
    
    dim1, dim2 = result.shape
    
    assert dim1 == dim2, "error proximity matrix not quadratic"
    assert dim1 == len(X), "error proximity matrix has wrong dimensions"
    
    assert np.diag(result).min() == 1.
    assert np.diag(result).max() == 1.


############################################
# imports
############################################

import sys
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.patches import Patch

from sklearn_extra.cluster import KMedoids

import fgclustering.utils as utils
import fgclustering.statistics as stats


############################################
# functions
############################################

def _plot_global_feature_importance(p_value_of_features, save):
    '''Plot global feature importance based on p-values given as input.

    :param p_value_of_features: dictionary where keys are names of features and values are p-values of these features
    :type p_value_of_features: dict
    :param save: Filename to save plot.
    :type save: str
    '''
    
    importance = p_value_of_features.copy()
    importance.pop('target')
    importance.pop('cluster')
    importance = pd.DataFrame(importance, index=[0])
    importance = pd.melt(importance)
    importance.sort_values(by='value', ascending=True, inplace=True)
    importance.value = 1 - importance.value

    plot = sns.barplot(data=importance, x='value', y='variable', color='darkgrey')
    plot.set_xlabel('importance')
    plot.set_ylabel('feature')
    plt.title('Global Feature Importance')
    plt.tight_layout()
    plt.show()

    if save is not None:
        plt.savefig('{}_global_feature_importance.png'.format(save), bbox_inches='tight', dpi = 300)
    

def _plot_local_feature_importance(X, bootstraps, save, num_cols):
    '''Plot local feature importance to show the importance of each feature for each cluster. 

    :param X: Feature matrix.
    :type X: pandas.DataFrame
    :param bootstraps: Number of bootstraps to be drawn for computation of p-value.
    :type bootstraps: int
    :param save: Filename to save plot.
    :type save: str
    :param num_cols: Number of plots in one row.
    :type num_cols: int
    '''
    
    importance = stats.get_feature_importance_clusterwise(X, bootstraps)
    num_features = len(importance)

    X_barplot = pd.melt(importance , ignore_index=False)
    X_barplot = X_barplot.rename_axis('feature').reset_index(level=0, inplace=False)
    X_barplot = X_barplot.sort_values('value',ascending=False)

    height = max(5,int(np.ceil(5*num_features/25)))
    num_cols = min(num_cols, len(importance.columns))

    plot = sns.FacetGrid(X_barplot, col='variable', sharey=False, col_wrap=num_cols,height=height)
    plot.map(sns.barplot, 'value', 'feature', color='darkgrey')
    plot.set_axis_labels('importance', 'feature')
    plot.set_titles(col_template="Cluster {col_name}")
    plt.suptitle('Feature Importance per Cluster')
    plt.tight_layout()
    plt.show()

    if save is not None:
        plt.savefig('{}_local_feature_importance.png'.format(save), bbox_inches='tight', dpi = 300)
    

def _plot_heatmap(X, method, save):
    '''Plot feature heatmap sorted by clusters, where features are filtered and ranked 
    with statistical tests (ANOVA for continuous featres, chi square for categorical features). 

    :param X: Feature matrix.
    :type X: pandas.DataFrame
    :param method: Model type of Random Forest model: classifier or regression.
    :type method: str
    :param save: Filename to save plot.
    :type save: str
    '''
    
    X_scaled = utils.scale_minmax(X)
    X_heatmap = pd.DataFrame(columns = X_scaled.columns)

    target_values_original = X['target']
    target_values_scaled = X_scaled['target']
    
    one_percent_of_number_of_samples = int(np.ceil(0.01*len(X)))

    for cluster in X_scaled.cluster.unique():
        X_heatmap = X_heatmap.append(X_scaled[X_scaled.cluster == cluster], ignore_index=True)
        X_heatmap = X_heatmap.append(pd.DataFrame(np.nan, 
                                                  index = np.arange(one_percent_of_number_of_samples), #blank lines which are 1% of num samples
                                                  columns = X_scaled.columns), ignore_index=True)
    X_heatmap = X_heatmap[:-5]
    X_heatmap.drop('cluster', axis=1, inplace=True)

    n_samples, n_features = X_heatmap.shape
    heatmap_ = np.zeros((n_features,n_samples,4))

    cmap_features = matplotlib.cm.get_cmap('coolwarm').copy()
    cmap_target = matplotlib.cm.get_cmap('viridis') #gist_ncar

    for feature in range(n_features):
        for sample in range(n_samples):
            if feature == 0:
                heatmap_[feature,sample,:] = cmap_target(X_heatmap.iloc[sample, feature])
            else:
                heatmap_[feature,sample,:] = cmap_features(X_heatmap.iloc[sample, feature])

    fig = plt.figure()
    img = plt.imshow(heatmap_, interpolation='none', aspect='auto')

    plt.title('Forest-Guided Clustering')
    plt.xticks([], [])
    plt.yticks(range(n_features), X_heatmap.columns)

    # remove bounding box
    for spine in plt.gca().spines.values():
        spine.set_visible(False)
    
    if method == "regression":
        norm = matplotlib.colors.Normalize(vmin=target_values_original.min(), vmax=target_values_original.max())
        cbar_target = plt.colorbar(matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap_target))
        cbar_target.set_label('target')
    else:
        legend_elements = [Patch(facecolor=cmap_target(tv_n), 
                                 edgecolor=cmap_target(tv_n),
                                 label=f'{tv_o}') for tv_n, tv_o in zip(target_values_scaled.unique(), target_values_original.unique())]
        ll = plt.legend(handles=legend_elements,
                        bbox_to_anchor=(0, 0), 
                        loc='upper left', 
                        ncol=min(6,len(legend_elements)),
                        title="target") 
    
    norm = matplotlib.colors.Normalize(vmin=0, vmax=1)
    cbar_features = plt.colorbar(matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap_features))
    cbar_features.set_label('standardized feature values')
    plt.show()
    
    if save is not None:
        plt.savefig('{}_heatmap.png'.format(save), bbox_inches='tight', dpi = 300)
    

def _plot_boxplots(X, save, num_cols):
    '''Plot feature boxplots divided by clusters, where features are filtered and ranked 
    with statistical tests (ANOVA for continuous featres, chi square for categorical features).

    :param X: Feature matrix.
    :type X: pandas.DataFrame
    :param save: Filename to save plot.
    :type save: str
    :param num_cols: Number of plots in one row.
    :type num_cols: int
    '''    
    
    target_and_features = X.columns[X.columns != 'cluster']
    X_boxplot = pd.melt(X, id_vars=['cluster'], value_vars=target_and_features)
    
    plot = sns.FacetGrid(X_boxplot, col='variable', height=3, sharey=False, col_wrap=num_cols)
    plot.map(sns.boxplot, 'cluster', 'value', color='darkgrey')
    plot.set_axis_labels('Cluster', 'Feature Value', clear_inner=False)
    plot.set_titles(col_template="Feature: {col_name}")
    plt.show()

    if save is not None:
        plt.savefig('{}_boxplots.png'.format(save), bbox_inches='tight', dpi = 300)
        



############################################
# imports
############################################

import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler, MinMaxScaler


############################################
# functions
############################################

def scale_standard(X):
    '''Feature Scaling with StandardScaler.

    :param X: Feature matrix.
    :type X: pandas.DataFrame
    :return: Standardized feature matrix.
    :rtype: pandas.DataFrame
    '''    
    X = X.copy()
    SCALE = StandardScaler()
    SCALE.fit(X)

    X_scale = pd.DataFrame(SCALE.transform(X))
    X_scale.columns = X.columns
    X_scale.reset_index(inplace=True,drop=True)

    return X_scale


def scale_minmax(X):
    '''Feature Scaling with MinMaxScaler.

    :param X: Feature matrix.
    :type X: pandas.DataFrame
    :return: Standardized feature matrix.
    :rtype: pandas.DataFrame
    ''' 
    X = X.copy()
    SCALE = MinMaxScaler()
    SCALE.fit(X)

    X_scale = pd.DataFrame(SCALE.transform(X))
    X_scale.columns = X.columns
    X_scale.reset_index(inplace=True,drop=True)

    return X_scale


def proximityMatrix(model, X, normalize=True):  
    '''Calculate proximity matrix of Random Forest model. 

    :param model: Trained Random Forest model.
    :type model: sklearn.ensemble
    :param X: Feature matrix.
    :type X: pandas.DataFrame
    :param normalize: Normalize proximity matrix by number of trees in the Random Forest, defaults to True.
    :type normalize: bool, optional
    :return: Proximity matrix of Random Forest model.
    :rtype: pandas.DataFrame
    '''
    terminals = model.apply(X)
    nTrees = terminals.shape[1]

    a = 0
    proxMat = 0

    for i in range(nTrees):
        a = terminals[:,i]
        proxMat += np.equal.outer(a, a)

    if normalize:
        proxMat = proxMat / nTrees

    return proxMat 


############################################
# imports
############################################

from sklearn_extra.cluster import KMedoids

import fgclustering.utils as utils
import fgclustering.optimizer as optimizer
import fgclustering.plotting as plotting
import fgclustering.statistics as stats

############################################
# Forest-guided Clustering
############################################


class FgClustering():
    '''Forest-Guided Clustering. 

    Computes a feature importance based on subgroups of instances that follow similar decision 
    rules within the Random Forest model.

    :param model: Trained Random Forest model.
    :type model: sklearn.ensemble
    :param data: Input data with feature matrix. 
        If target_column is a string it has to be a column in the data.
    :type data: pandas.DataFrame
    :param target_column: Name of target column or target values as numpy array.
    :type target_column: str or numpy.ndarray
    :param random_state: [description], defaults to 42
    :type random_state: int, optional
    :raises ValueError: error raised if Random Forest model is not a 
        sklearn.ensemble.RandomForestClassifier or sklearn.ensemble.RandomForestRegressor object
    '''
    def __init__(self, model, data, target_column, random_state = 42):
        self.random_state = random_state

        # check if random forest is regressor or classifier
        is_regressor = 'RandomForestRegressor' in str(type(model))
        is_classifier = 'RandomForestClassifier' in str(type(model))
        
        if is_regressor is True:
            self.method = "regression"
            print("Interpreting RandomForestRegressor")
        elif is_classifier is True:
            self.method = "classifier"
            print("Interpreting RandomForestClassifier")
        else:
            raise ValueError(f'Do not recognize {str(type(model))}. Can only work with sklearn RandomForestRegressor or RandomForestClassifier.')

        if type(target_column)==str:
            self.y = data.loc[:,target_column]
            self.X = data.drop(columns=[target_column])
        else:
            self.y = target_column
            self.X = data
        
        self.proximity_matrix = utils.proximityMatrix(model, self.X.to_numpy())
        self.distance_matrix = 1 - self.proximity_matrix
        self.k = None
        self.cluster_labels = None

    def run(self, number_of_clusters = None, max_K = 8, bootstraps_JI = 300, max_iter_clustering = 500, discart_value_JI = 0.6):
        '''Runs the forest-guided clustering model. The optimal number of clusters for a k-medoids clustering is computed, 
        based on the distance matrix computed from the Random Forest proximity matrix.

        :param number_of_clusters: Number of clusters for the k-medoids clustering. 
            Leave None if number of clusters should be optimized, defaults to None
        :type number_of_clusters: int, optional
        :param max_K: Maximum number of clusters for cluster score computation, defaults to 8
        :type max_K: int, optional
        :param bootstraps_JI: Number of bootstraps to compute the Jaccard Index, defaults to 300
        :type bootstraps_JI: int, optional  
        :param max_iter_clustering: Number of iterations for k-medoids clustering, defaults to 500
        :type max_iter_clustering: int, optional
        :param discart_value_JI: Minimum Jaccard Index for cluster stability, defaults to 0.6
        :type discart_value_JI: float, optional
        '''
        if number_of_clusters is None:
            self.k = optimizer.optimizeK(self.distance_matrix, 
                                    self.y.to_numpy(), 
                                    max_K, 
                                    bootstraps_JI, 
                                    max_iter_clustering, 
                                    discart_value_JI, 
                                    self.method, 
                                    self.random_state)
            print(f"Optimal number of cluster is: {self.k}")
        else:
            self.k = number_of_clusters
            print(f"Use {self.k} as number of cluster")

        self.cluster_labels = KMedoids(n_clusters=self.k, random_state=self.random_state).fit(self.distance_matrix).labels_
        self._X_ranked, self.p_value_of_features = stats.calculate_global_feature_importance(self.X, self.y, self.cluster_labels)

    def plot_global_feature_importance(self, save = None):
        '''Plot global feature importance based on p-values given as input, the p-values are computed using an Anova (for continuous
        variable) or a Chi-Square (for categorical variables) test. The features importance is defined by 1-p_value.
        
        :param save: Filename to save plot.
        :type save: str

        '''
        plotting._plot_global_feature_importance(self.p_value_of_features, save)

    def plot_local_feature_importance(self, bootstraps_p_value = 1000, thr_pvalue = 0.01, save = None, num_cols = 4):
        '''Plot local feature importance to show the importance of each feature for each cluster, 
        measured by variance and impurity of the feature within the cluster, i.e. the higher 
        the feature importance, the lower the feature variance / impurity within the cluster.

        :param bootstraps_p_value: Number of bootstraps to compute the p-value of feature importance, defaults to 1000
        :type bootstraps_p_value: int, optional
        :param thr_pvalue: P-value threshold for feature filtering, defaults to 0.01
        :type thr_pvalue: float, optional
        :param save: Filename to save plot, if None the figure is not saved, defaults to None
        :type save: str, optional
        :param num_cols: Number of plots in one row, defaults to 4.
        :type num_cols: int, optional
        '''
        # drop insignificant values
        X_ranked = self._X_ranked.copy()
        for column in X_ranked.columns:
            if self.p_value_of_features[column] > thr_pvalue:
                X_ranked.drop(column, axis  = 1, inplace=True) 

        plotting._plot_local_feature_importance(X_ranked, bootstraps_p_value, save, num_cols)

    def plot_heatmap(self, thr_pvalue = 0.01, save = None):
        '''Plot feature heatmap sorted by clusters, where features are filtered and ranked 
        with statistical tests (ANOVA for continuous featres, chi square for categorical features). 

        :param thr_pvalue: P-value threshold for feature filtering, defaults to 0.01
        :type thr_pvalue: float, optional
        :param save: Filename to save plot, if None the figure is not saved, defaults to None
        :type save: str, optional
        '''
        # drop insignificant values
        X_ranked = self._X_ranked.copy()
        for column in X_ranked.columns:
            if self.p_value_of_features[column] > thr_pvalue:
                X_ranked.drop(column, axis  = 1, inplace=True)    

        plotting._plot_heatmap(X_ranked, self.method, save)

    def plot_boxplots(self, thr_pvalue = 0.01, save = None, num_cols = 6):
        '''Plot feature boxplots divided by clusters, where features are filtered and ranked 
        with statistical tests (ANOVA for continuous featres, chi square for categorical features).

        :param thr_pvalue: P-value threshold for feature filtering, defaults to 0.01
        :type thr_pvalue: float, optional
        :param save: Filename to save plot, if None the figure is not saved, defaults to None
        :type save: str, optional
        :param num_cols: Number of plots in one row, defaults to 6.
        :type num_cols: int, optional
        '''
        # drop insignificant values
        X_ranked = self._X_ranked.copy()
        for column in X_ranked.columns:
            if self.p_value_of_features[column] > thr_pvalue:
                X_ranked.drop(column, axis  = 1, inplace=True) 

        plotting._plot_boxplots(X_ranked, save, num_cols)



############################################
# imports
############################################

import joblib
import numpy as np
import pandas as pd

from sklearn_extra.cluster import KMedoids

from fgclustering.utils import *
from fgclustering.statistics import compute_balanced_average_impurity, compute_total_within_cluster_variation, calculate_global_feature_importance, get_feature_importance_clusterwise


############################################
# Tests
############################################

def test_compute_balanced_average_impurity():
    
    expected_result = 0.

    categorical_values = pd.Series([0,0,0,0,0,1,1,1,1,1])
    cluster_labels = np.array([0,0,0,0,0,1,1,1,1,1])
    
    result = compute_balanced_average_impurity(categorical_values, cluster_labels)
    
    assert expected_result == result


def test_compute_total_within_cluster_variation():
    
    expected_result = 0.

    continuous_values = pd.Series([0,0,0,0,0,1,1,1,1,1])
    cluster_labels = np.array([0,0,0,0,0,1,1,1,1,1])
    
    result = compute_total_within_cluster_variation(continuous_values, cluster_labels)
    
    assert expected_result == result


def test_calculate_global_feature_importance():
    
    # test if anova test filters out features 1 and 2 which are the same in both clusters and 
    # leaves features 3 and 4 which are clearly different in both clusters
    
    X = pd.DataFrame.from_dict({'col_1': [1,1,1,1,1,0.9], 'col_2': [1,1,1,1,0.9,0.5], 'col_3': [1,1,1,0,0,1], 'col_4': [0,0,0,1,1,1]})
    y = pd.Series([0,0,0,0,0,0])
    cluster_labels = np.array([0,0,0,1,1,1])
    
    X_ranked, p_value_of_features = calculate_global_feature_importance(X, y, cluster_labels)
    
    X_ranked.drop('cluster', axis  = 1, inplace=True)
    assert list(X_ranked.columns) == ['target', 'col_4', 'col_3', 'col_2', 'col_1']


def test_get_feature_importance_clusterwise():

    k = 3
    random_state = 42
    thr_pvalue = 0.001
    bootstraps = 100

    data_iris = pd.read_csv('./data/data_iris.csv')
    model = joblib.load(open('./data/random_forest_iris.joblib', 'rb'))
    X = data_iris.drop(columns=['target'])
    y = data_iris.loc[:,'target'].to_numpy()

    distanceMatrix = 1 - proximityMatrix(model, X.to_numpy())

    cluster_labels = KMedoids(n_clusters=k, random_state=random_state).fit(distanceMatrix).labels_
    X_ranked, p_value_of_features = calculate_global_feature_importance(X, y, cluster_labels)
    for column in X.columns:
        if p_value_of_features[column] > thr_pvalue:
            X.drop(column, axis  = 1, inplace=True)    

    importance = get_feature_importance_clusterwise(X_ranked, bootstraps)
    
    assert sum(importance < 10) == 3






############################################
# imports
############################################

import sys
import numpy as np
import pandas as pd

from bisect import bisect
from scipy.stats import f_oneway, chisquare
from statsmodels.stats import multitest
from sklearn.utils import resample


############################################
# functions
############################################


def compute_balanced_average_impurity(categorical_values, cluster_labels):
    '''Compute balanced average impurity as score for categorical values in a clustering. 
    Impurity score is an Gini Coefficient of the classes within each cluster. 
    The class sizes are balanced by rescaling with the inverse size of the class in the overall dataset.

    :param categorical_values: Values of categorical feature / target.
    :type categorical_values: pandas.Series
    :param cluster_labels: Cluster labels for each value.
    :type cluster_labels: numpy.ndarray
    :return: Impurity score.
    :rtype: float
    '''
    #compute the number of datapoints for each class to use it then for rescaling of the 
    #class sizes within each cluster --> rescaling with inverse class size
    rescaling_factor = {class_: 1/sum(categorical_values==class_) for class_ in np.unique(categorical_values)} 
    balanced_impurities = []

    for cluster in np.unique(cluster_labels):
        categorical_values_cluster = categorical_values[cluster_labels == cluster]
        
        #compute balanced class probabilities (rescaled with overall class size)
        class_probabilities_unnormalized = [sum(categorical_values_cluster==class_)*rescaling_factor[class_] for class_ in np.unique(categorical_values)]
        class_probabilities_unnormalized = np.array(class_probabilities_unnormalized)
        normalization_factor = class_probabilities_unnormalized.sum()
        class_probabilities = class_probabilities_unnormalized / normalization_factor
        
        #compute (balanced) gini impurity
        gini_impurity = 1 - np.sum(class_probabilities**2)
        balanced_impurities.append(gini_impurity)
    
    score = np.mean(balanced_impurities)
    
    return score


def compute_total_within_cluster_variation(continuous_values, cluster_labels):
    '''Compute total within cluster variation as score for continuous values in a clustering.

    :param continuous_values: Values of continuous feature / target.
    :type continuous_values: pandas.Series
    :param cluster_labels: Cluster labels for each value.
    :type cluster_labels: numpy.ndarray
    :return: Within cluster variation score.
    :rtype: float
    '''
    score = 0
    for cluster in np.unique(cluster_labels):
        continuous_values_cluster = continuous_values[cluster_labels == cluster]
        score += np.var(continuous_values_cluster)*len(continuous_values_cluster)
        
    return score


def _anova_test(list_of_df):
    '''Perform one way ANOVA test on continuous features.

    :param list_of_df: List of dataframes, where each dataframe contains 
        the feature values for one cluster.
    :type list_of_df: list
    :return: P-value of ANOVA test.
    :rtype: float
    '''
    anova = f_oneway(*list_of_df)
    return anova.pvalue


def _chisquare_test(df, list_of_df):
    '''Perform chi square test on categorical features.

    :param df: Dataframe with feature and cluster.
    :type df: pandas.DataFrame
    :param list_of_df: List of dataframes, where each dataframe contains 
        the feature values for one cluster.
    :type list_of_df: list
    :return: P-value of chi square test.
    :rtype: float
    '''
    cat_vals = df.feature.unique()
    count_global = np.array([(df.feature == cat_val).sum() for cat_val in cat_vals])
    count_global = count_global / count_global.sum()

    p_values = []
    for df_ in list_of_df:
        counts_clusters = np.array([(df_ == cat_val).sum() for cat_val in cat_vals])
        number_datapoints_in_cluster = counts_clusters.sum()
        p_values.append(chisquare(counts_clusters, f_exp=count_global*number_datapoints_in_cluster).pvalue)

    _, p_values = multitest.fdrcorrection(p_values)
    return min(p_values)


def _rank_features(X, y, p_value_of_features):
    '''Rank features by lowest p-value.

    :param X: Feature matrix.
    :type X: pandas.DataFrame
    :param y: Target column.
    :type y: pandas.Series  
    :param p_value_of_features: Computed p-values of all features.
    :type p_value_of_features: dict
    :return: Ranked feature matrix.
    :rtype: pandas.DataFrame
    '''
    X_ranked = X.copy()
    X_ranked['target'] = y
    p_value_of_features['target'] = -1
    p_value_of_features['cluster'] = -1
    
    # sort features by p-value
    features_sorted = [k for k, v in sorted(p_value_of_features.items(), key=lambda item: item[1])]
    X_ranked = X_ranked.reindex(features_sorted, axis=1)

    return X_ranked


def _sort_clusters_by_target(X_ranked):
    '''Sort clusters by mean target values in clusters. 

    :param X_ranked: Filtered and ranked feature matrix.
    :type X_ranked: pandas.DataFrame
    :return: Filtered and ranked feature matrix with ordered clusters.
    :rtype: pandas.DataFrame
    '''
    means = X_ranked.groupby(['cluster']).mean().sort_values(by='target',ascending=True)
    means['target'] = range(means.shape[0])
    mapping = dict(means['target'])
    mapping = dict(sorted(mapping.items(), key=lambda item: item[1]))
    X_ranked = X_ranked.replace({'cluster': mapping})
    X_ranked['cluster'] = pd.Categorical(X_ranked['cluster'], sorted(X_ranked['cluster'].unique()))

    return X_ranked


def calculate_global_feature_importance(X, y, cluster_labels):
    '''Calculate global feature importance for each feature. 
    The higher the importance for a feature, the lower the p-value obtained by 
    an ANOVA (continuous feature) or chi-square (categorical feature) test.

    :param X: Feature matrix.
    :type X: pandas.DataFrame
    :param y: Target column.
    :type y: pandas.Series
    :param cluster_labels: Clustering labels.
    :type cluster_labels: numpy.ndarray
    :return: Feature matrix ranked by p-value of statistical test and 
        dictionary with computed p-values of all features.
    :rtype: pandas.DataFrame and dict
    ''' 
    X['cluster'] = cluster_labels
    p_value_of_features = dict()

    # statistical test for each feature
    for feature in X.columns:
        assert X[feature].astype("object").dtype != X[feature].dtype, "Feature is of type object. Please reformat to category or numeric type."

        df = pd.DataFrame({'cluster': X['cluster'], 'feature': X[feature]})
        list_of_df = [df.feature[df.cluster == cluster] for cluster in set(df.cluster)]
        
        if isinstance(X[feature].dtype, pd.api.types.CategoricalDtype):
            chisquare_p_value = _chisquare_test(df, list_of_df)
            p_value_of_features[feature] = chisquare_p_value
            
        else:  
            anova_p_value = _anova_test(list_of_df)
            p_value_of_features[feature] = anova_p_value

    X_ranked = _rank_features(X, y, p_value_of_features)
    X_ranked = _sort_clusters_by_target(X_ranked)
    X_ranked.sort_values(by=['cluster','target'], axis=0, inplace=True)
    
    return X_ranked, p_value_of_features


def _calculate_p_value_categorical(X_feature_cluster, X_feature, cluster, cluster_size, bootstraps):
    '''Calculate bootstrapped p-value for categorical features to
    determine the importance of the feature for a certain cluster. 
    The lower the bootstrapped p-value, the lower the impurity of the 
    feature in the respective cluster.

    :param X_feature_cluster: Categorical feature values in cluster.
    :type X_feature_cluster: pandas.Series
    :param X_feature: Categorical feature values.
    :type X_feature: pandas.Series
    :param cluster: Cluster number.
    :type cluster: int
    :param cluster_size: Size of cluster, i.e. number of data points in cluster.
    :type cluster_size: int
    :param bootstraps: Number of bootstraps to be drawn for computation of p-value.
    :type bootstraps: int
    :return: Bootstrapped p-value for categorical feature.
    :rtype: float
    '''
    cluster_label = [cluster]*cluster_size
    X_feature_cluster_impurity = compute_balanced_average_impurity(X_feature_cluster, cluster_label)

    bootstrapped_impurity = list()
    for b in range(bootstraps):
        bootstrapped_X_feature = resample(X_feature, replace = True, n_samples = cluster_size)
        bootstrapped_impurity.append(compute_balanced_average_impurity(bootstrapped_X_feature, cluster_label))
        
    bootstrapped_impurity = sorted(bootstrapped_impurity)
    p_value = bisect(bootstrapped_impurity, X_feature_cluster_impurity) / bootstraps
    return p_value

    
def _calculate_p_value_continuous(X_feature_cluster, X_feature, cluster_size, bootstraps):
    '''Calculate bootstrapped p-value for continuous features to
    determine the importance of the feature for a certain cluster. 
    The lower the bootstrapped p-value, the lower the variance of the 
    feature in the respective cluster.

    :param X_feature_cluster: Continuous feature values in cluster.
    :type X_feature_cluster: pandas.Series
    :param X_feature: Continuous feature values.
    :type X_feature: pandas.Series
    :param cluster_size: Size of cluster, i.e. number of data points in cluster.
    :type cluster_size: int
    :param bootstraps: Number of bootstraps to be drawn for computation of p-value.
    :type bootstraps: int
    :return: Bootstrapped p-value for continuous feature.
    :rtype: float
    '''
    X_feature_cluster_var = X_feature_cluster.var()

    bootstrapped_var = list()
    for b in range(bootstraps):
        bootstrapped_X_feature = resample(X_feature, replace = True, n_samples = cluster_size)
        bootstrapped_var.append(bootstrapped_X_feature.var())
        
    bootstrapped_var = sorted(bootstrapped_var)
    p_value = bisect(bootstrapped_var, X_feature_cluster_var) / bootstraps
    return p_value
    

def get_feature_importance_clusterwise(X, bootstraps, epsilon = sys.float_info.min):
    '''Calculate local importance of each feature within each cluster. 
    The higher the importance for a feature, the lower the variance (continuous feature) 
    or impurity (categorical feature) of that feature within the cluster.

    :param X: Feature matrix.
    :type X: pandas.DataFrame
    :param bootstraps: Number of bootstraps to be drawn for computation of p-value.
    :type bootstraps: int
    :param epsilon: Small value for log calculation to avoind log(0), defaults to sys.float_info.min
    :type epsilon: float, optional
    :return: Importance matrix with feature importance per cluster.
    :rtype: pandas.DataFrame
    '''
    X = X.copy()
    clusters = X['cluster']
    clusters_size = clusters.value_counts()
    X.drop(['cluster', 'target'], axis=1, inplace=True)

    features = X.columns.tolist()
    importance = pd.DataFrame(columns=clusters.unique(), index=features)
    
    X_categorical = X.select_dtypes(include=['category'])
    X_numeric = X.select_dtypes(exclude=['category'])

    for feature in X_categorical.columns:
        for cluster in clusters.unique():
            X_feature_cluster = X_categorical.loc[clusters == cluster, feature]
            X_feature = X_categorical[feature]
            importance.loc[feature,cluster] = 1 - (_calculate_p_value_categorical(X_feature_cluster, X_feature, cluster, clusters_size.loc[cluster], bootstraps) + epsilon)

    for feature in X_numeric.columns:
        X_numeric.loc[:,feature] = X_numeric[feature]
        for cluster in clusters.unique():
            X_feature_cluster = X_numeric.loc[clusters == cluster, feature]
            X_feature = X_numeric[feature]
            importance.loc[feature,cluster] = 1 - (_calculate_p_value_continuous(X_feature_cluster, X_feature, clusters_size.loc[cluster], bootstraps) + epsilon)
    
    return importance


from distutils.core import setup
import setuptools

setup(
    name='fgclustering',
    version='0.2',
    url='https://github.com/HelmholtzAI-Consultants-Munich/forest_guided_clustering',
    author='Lisa Barros de Andrade e Sousa, Dominik Thalmeier',
    author_email='lisa.barros.andrade.sousa@gmail.com, dominikthalmeier@googlemail.com',
    packages=setuptools.find_packages(),
    install_requires=['pandas','numpy','matplotlib','seaborn','sklearn','scikit-learn-extra','scipy','tqdm', 'statsmodels','Sphinx==4.2.0'],
    long_description=open('README.rst').read(),
)



############################################
# imports
############################################

import joblib
import numpy as np
import pandas as pd

from sklearn_extra.cluster import KMedoids

from fgclustering.utils import proximityMatrix
from fgclustering.optimizer import optimizeK, _compute_stability_indices, _translate_cluster_labels_to_dictionary_of_index_sets_per_cluster


############################################
# Tests
############################################

def test_optimizeK():
    
    bootstraps = 30
    max_iter_clustering = 100
    discart_value = 0.6
    random_state = 42

    ### test classification
    max_K = 4
    method = "classifier"

    data_breast_cancer = pd.read_csv('./data/data_breast_cancer.csv')
    model = joblib.load(open('./data/random_forest_breat_cancer.joblib', 'rb'))
    X = data_breast_cancer.drop(columns=['target']).to_numpy()
    y = data_breast_cancer.loc[:,'target'].to_numpy()
    
    distance_matrix = 1 - proximityMatrix(model, X)
    result = optimizeK(distance_matrix, y, max_K, bootstraps, max_iter_clustering, discart_value, method, random_state)
    
    assert result == 2, "Error optimal number of Clusters for breast cancer test case is not equal 2"

    ### test regression
    max_K = 7
    method = "regression"

    data_boston = pd.read_csv('./data/data_boston.csv')
    model = joblib.load(open('./data/random_forest_boston.joblib', 'rb'))
    X = data_boston.drop(columns=['target']).to_numpy()
    y = data_boston.loc[:,'target'].to_numpy()
    
    distance_matrix = 1 - proximityMatrix(model, X)
    result = optimizeK(distance_matrix, y, max_K, bootstraps, max_iter_clustering, discart_value, method, random_state)

    assert result == 5 or result == 6, "Error optimal number of Clusters for boston test case is not equal 5 or 6" 


def test_compute_stability_indices():
    
    distance_matrix = np.kron(np.eye(3,dtype=int),np.ones([10,10]))
    bootstraps = 10
    random_state = 42
    
    #test 1: test if 3 different clusters are found and have maximal stability
    cluster_method = lambda X: KMedoids(n_clusters=3, random_state=42, init = 'build', method = "pam", max_iter=100).fit(X).labels_
    
    result = _compute_stability_indices(distance_matrix, cluster_method, bootstraps, random_state)
    
    assert result[0] == 1., "Clusters that should be stable are found to be unstable"
    assert result[1] == 1., "Clusters that should be stable are found to be unstable"
    assert result[2] == 1., "Clusters that should be stable are found to be unstable"
    

    #test 2: test if 3 different clusters are found and have maximal stability
    cluster_method = lambda X: KMedoids(n_clusters=2, random_state=42, init = 'build', method = "pam", max_iter=100).fit(X).labels_
    
    result = _compute_stability_indices(distance_matrix, cluster_method, bootstraps, random_state)
    
    assert min(result[0], result[1]) < 1., "Clusters that should be unstable are found to be stable"


def test_translate_cluster_labels_to_dictionary_of_index_sets_per_cluster():
    
    labels = [1,2,3,2,1]
    expected_output = {1: set([0,4]), 2: set([1,3]), 3: set([2])}
    
    output = _translate_cluster_labels_to_dictionary_of_index_sets_per_cluster(labels, mapping = False)

    assert output == expected_output
    
    expected_output = {1: set([10,14]), 2: set([11,13]), 3: set([12])}
    output = _translate_cluster_labels_to_dictionary_of_index_sets_per_cluster(labels, mapping = {0:10,1:11,2:12,3:13,4:14,5:15})
    
    assert output == expected_output
    
    
